/* SPDX-License-Identifier: GPL-2.0 */

#ifndef _ASM_X86_NOSPEC_BRANCH_H_
#define _ASM_X86_NOSPEC_BRANCH_H_

#include <linux/static_key.h>

#include <asm/alternative.h>
#include <asm/alternative-asm.h>
#include <asm/cpufeature.h>
#include <asm/msr-index.h>

#ifdef __ASSEMBLY__

/*
 * MDS_USER_CLEAR_CPU_BUFFERS macro is the assembly equivalent of
 * mds_user_clear_cpu_buffers(). Like the C version, the __KERNEL_DS
 * is used for verw.
 */
.macro MDS_USER_CLEAR_CPU_BUFFERS
#ifdef CONFIG_JUMP_LABEL
	STATIC_JUMP_IF_TRUE .Lmds_clear_\@, mds_user_clear
	jmp	.Lmds_skip_clear_\@
.Lmds_clear_\@:
#endif
#ifdef CONFIG_X86_64
	verw	mds_clear_cpu_buffers_ds(%rip)
#else
	verw	mds_clear_cpu_buffers_ds
#endif
.Lmds_skip_clear_\@:
.endm

#else /* __ASSEMBLY__ */

/* The IBPB runtime control knob */
extern unsigned int ibpb_enabled;
int set_ibpb_enabled(unsigned int);

/* The IBRS runtime control knob */
extern unsigned int ibrs_enabled;
int set_ibrs_enabled(unsigned int);

static __always_inline
void alternative_msr_write(unsigned int msr, u64 val, unsigned int feature)
{
	asm volatile(ALTERNATIVE("", "wrmsr", %c[feature])
		: : "c" (msr),
		    "a" ((u32)val),
		    "d" ((u32)(val >> 32)),
		    [feature] "i" (feature)
		: "memory");
}

static inline void indirect_branch_prediction_barrier(void)
{
	u64 val = PRED_CMD_IBPB;

	if (ibpb_enabled)
		alternative_msr_write(MSR_IA32_PRED_CMD, val,
				      X86_FEATURE_USE_IBPB);
}

/* The Intel SPEC CTRL MSR base value cache */
extern u64 x86_spec_ctrl_base;

/*
 * With retpoline, we must use IBRS to restrict branch prediction
 * before calling into firmware.
 *
 * (Implemented as CPP macros due to header hell.)
 */
#define firmware_restrict_branch_speculation_start()			\
do {									\
	u64 val = x86_spec_ctrl_base | SPEC_CTRL_IBRS;			\
									\
	preempt_disable();						\
	alternative_msr_write(MSR_IA32_SPEC_CTRL, val,			\
			      X86_FEATURE_USE_IBRS_FW);			\
} while (0)

#define firmware_restrict_branch_speculation_end()			\
do {									\
	u64 val = x86_spec_ctrl_base;					\
									\
	alternative_msr_write(MSR_IA32_SPEC_CTRL, val,			\
			      X86_FEATURE_USE_IBRS_FW);			\
	preempt_enable();						\
} while (0)

#define ubuntu_restrict_branch_speculation_start()			\
do {									\
	u64 val = x86_spec_ctrl_base | SPEC_CTRL_IBRS;			\
									\
	if (ibrs_enabled)						\
		native_wrmsrl(MSR_IA32_SPEC_CTRL, val);			\
} while (0)

#define ubuntu_restrict_branch_speculation_end()			\
do {									\
	u64 val = x86_spec_ctrl_base;					\
									\
	if (ibrs_enabled)						\
		native_wrmsrl(MSR_IA32_SPEC_CTRL, val);			\
} while (0)

DECLARE_JUMP_LABEL_KEY_FALSE(mds_user_clear);
DECLARE_JUMP_LABEL_KEY_FALSE(mds_idle_clear);

#include <asm/segment.h>

/**
 * mds_clear_cpu_buffers - Mitigation for MDS and TAA vulnerability
 *
 * This uses the otherwise unused and obsolete VERW instruction in
 * combination with microcode which triggers a CPU buffer flush when the
 * instruction is executed.
 */
static inline void mds_clear_cpu_buffers(void)
{
	static const u16 ds = __KERNEL_DS;

	/*
	 * Has to be the memory-operand variant because only that
	 * guarantees the CPU buffer flush functionality according to
	 * documentation. The register-operand variant does not.
	 * Works with any segment selector, but a valid writable
	 * data segment is the fastest variant.
	 *
	 * "cc" clobber is required because VERW modifies ZF.
	 */
	asm volatile("verw %[ds]" : : [ds] "m" (ds) : "cc");
}

/**
 * mds_user_clear_cpu_buffers - Mitigation for MDS and TAA vulnerability
 *
 * Clear CPU buffers if the corresponding static key is enabled
 */
static inline void mds_user_clear_cpu_buffers(void)
{
	if (static_branch_likely_init_false(&mds_user_clear))
		mds_clear_cpu_buffers();
}

/**
 * mds_idle_clear_cpu_buffers - Mitigation for MDS vulnerability
 *
 * Clear CPU buffers if the corresponding static key is enabled
 */
static inline void mds_idle_clear_cpu_buffers(void)
{
	if (static_branch_likely_init_false(&mds_idle_clear))
		mds_clear_cpu_buffers();
}

#endif /* __ASSEMBLY__ */
#endif /* _ASM_X86_NOSPEC_BRANCH_H_ */
