#ifndef _LINUX_JUMP_LABEL_H
#define _LINUX_JUMP_LABEL_H

#if defined(CC_HAVE_ASM_GOTO) && defined(CONFIG_JUMP_LABEL)
# define HAVE_JUMP_LABEL
#endif

#ifndef __ASSEMBLY__

#include <linux/types.h>
#include <linux/compiler.h>
#include <linux/bug.h>

#ifdef HAVE_JUMP_LABEL

struct jump_label_key {
	atomic_t enabled;
	struct jump_entry *entries;
#ifdef CONFIG_MODULES
	struct jump_label_mod *next;
#endif
};

#else
struct jump_label_key {
	atomic_t enabled;
};
#endif	/* HAVE_JUMP_LABEL */
#endif /* __ASSEMBLY__ */

#ifdef HAVE_JUMP_LABEL
#include <asm/jump_label.h>
#endif

#ifndef __ASSEMBLY__

enum jump_label_type {
	JUMP_LABEL_DISABLE = 0,
	JUMP_LABEL_ENABLE,
};

struct module;

#include <linux/atomic.h>

#ifdef HAVE_JUMP_LABEL

static inline int jump_label_key_count(struct jump_label_key *key)
{
	/*
	 * -1 means the first jump_label_inc() is in progress.
	 *  jump_label_key_enabled() must return true, so return 1 here.
	 */
	int n = atomic_read(&key->enabled);
	return n >= 0 ? n : 1;
}

#ifdef CONFIG_MODULES
#define JUMP_LABEL_INIT {ATOMIC_INIT(0), NULL, NULL}
#else
#define JUMP_LABEL_INIT {ATOMIC_INIT(0), NULL}
#endif

static __always_inline bool static_branch(struct jump_label_key *key)
{
	return arch_static_branch(key);
}

#define jump_label_key_false(key)	static_branch(key)

static __always_inline bool jump_label_key_true(struct jump_label_key *key)
{
        return !jump_label_key_false(key);
}

extern struct jump_entry __start___jump_table[];
extern struct jump_entry __stop___jump_table[];

extern void jump_label_init(void);
extern void jump_label_lock(void);
extern void jump_label_unlock(void);
extern void arch_jump_label_transform(struct jump_entry *entry,
				      enum jump_label_type type);
extern void arch_jump_label_transform_static(struct jump_entry *entry,
					     enum jump_label_type type);
extern int jump_label_text_reserved(void *start, void *end);
extern void jump_label_inc(struct jump_label_key *key);
extern void jump_label_dec(struct jump_label_key *key);
extern bool jump_label_enabled(struct jump_label_key *key);
extern void jump_label_apply_nops(struct module *mod);

#define JUMP_LABEL_KEY_INIT_TRUE ((struct jump_label_key)	\
	{ .enabled = { 1 }, .entries = (void *)1 })
#define JUMP_LABEL_KEY_INIT_FALSE ((struct jump_label_key)	\
	{ .enabled = { 0 }, .entries = (void *)0 })

#else  /* !HAVE_JUMP_LABEL */

#define JUMP_LABEL_INIT {ATOMIC_INIT(0)}

static inline int jump_label_key_count(struct jump_label_key *key)
{
	return atomic_read(&key->enabled);
}

static __always_inline void jump_label_init(void)
{
}

static __always_inline bool static_branch(struct jump_label_key *key)
{
	if (unlikely(jump_label_key_count(key) > 0))
		return true;
	return false;
}

#define jump_label_key_false(key)	static_branch(key)

static __always_inline bool jump_label_key_true(struct jump_label_key *key)
{
	if (likely(jump_label_key_count(key) > 0))
		return true;
	return false;
}

static inline void jump_label_inc(struct jump_label_key *key)
{
	atomic_inc(&key->enabled);
}

static inline void jump_label_dec(struct jump_label_key *key)
{
	atomic_dec(&key->enabled);
}

static inline int jump_label_text_reserved(void *start, void *end)
{
	return 0;
}

static inline void jump_label_lock(void) {}
static inline void jump_label_unlock(void) {}

static inline bool jump_label_enabled(struct jump_label_key *key)
{
	return !!atomic_read(&key->enabled);
}

static inline int jump_label_apply_nops(struct module *mod)
{
	return 0;
}

#define JUMP_LABEL_KEY_INIT_TRUE ((struct jump_label_key)	\
	{ .enabled = ATOMIC_INIT(1) })
#define JUMP_LABEL_KEY_INIT_FALSE ((struct jump_label_key)	\
	{ .enabled = ATOMIC_INIT(0) })

#endif	/* HAVE_JUMP_LABEL */

static inline bool jump_label_key_enabled(struct jump_label_key *key)
{
	return jump_label_key_count(key) > 0;
}

static inline void jump_label_key_enable(struct jump_label_key *key)
{
	int count = jump_label_key_count(key);

	WARN_ON_ONCE(count < 0 || count > 1);

	if (!count)
		jump_label_inc(key);
}

static inline void jump_label_key_disable(struct jump_label_key *key)
{
	int count = jump_label_key_count(key);

	WARN_ON_ONCE(count < 0 || count > 1);

	if (count)
		jump_label_dec(key);
}

#define DEFINE_JUMP_LABEL_KEY_TRUE(name)			\
	struct jump_label_key name = JUMP_LABEL_KEY_INIT_TRUE

#define DECLARE_JUMP_LABEL_KEY_TRUE(name)	\
	extern struct jump_label_key name

#define DEFINE_JUMP_LABEL_KEY_FALSE(name)			\
	struct jump_label_key name = JUMP_LABEL_KEY_INIT_FALSE

#define DECLARE_JUMP_LABEL_KEY_FALSE(name)	\
	extern struct jump_label_key name

#ifdef HAVE_JUMP_LABEL

/*
 * juergh: Helper #defines to (hopefully) prevent mistakes when using the
 * convoluted static key API. The current API only allows for the two
 * combinations below which execute without taking an uncoditional branch.
 * Note, that a new API was introduced later with commit 11276d5306b8
 * ("locking/static_keys: Add a new static_key interface").
 */
#define static_branch_unlikely_init_false(x)	jump_label_key_false(x)
#define static_branch_likely_init_true(x)	jump_label_key_true(x)

/*
 * juergh: Add more helper #defines to work around the static key API.
 * Note, that the following are 'unoptimized' #defines due to the limitation
 * of the current/old static key API.
 */
#define static_branch_likely_init_false(x)	likely(jump_label_key_enabled(x))
#define static_branch_unlikely_init_true(x)	unlikely(jump_label_key_enabled(x))

#else  /* !HAVE_JUMP_LABEL */

#define static_branch_likely_init_false(x)	likely(jump_label_key_enabled(x))
#define static_branch_likely_init_true(x)	likely(jump_label_key_enabled(x))
#define static_branch_unlikely_init_false(x)	unlikely(jump_label_key_enabled(x))
#define static_branch_unlikely_init_true(x)	unlikely(jump_label_key_enabled(x))

#endif

/*
 * Advanced usage; refcount, branch is enabled when: count != 0
 */

#define static_branch_inc(x)		jump_label_inc(x)
#define static_branch_dec(x)		jump_label_dec(x)

/*
 * Normal usage; boolean enable/disable.
 */

#define static_branch_enable(x)		jump_label_key_enable(x)
#define static_branch_disable(x)	jump_label_key_disable(x)

#endif	/* _LINUX_JUMP_LABEL_H */

#endif /* __ASSEMBLY__ */
